#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/power_supply.h>
#include <linux/broadcom/pmu_chip.h>
#include <linux/broadcom/bcm_adc.h>
#include <asm/arch/bcm_gen_battery.h>
#include <linux/delay.h>

#define PFX "bcm4760_battery"

struct bcm_ll_adc_req {
	bcm_adc_request_t	adc;	/* ADC request structure for request function */
	unsigned short		sample;	/* sample response from ADC integration */
	struct completion	cmp;	/* completion structure for this request */
};

static DEFINE_MUTEX(bcm_bat_mutex);
static DEFINE_MUTEX(bcm_vmbat_mutex);

#if defined(CONFIG_PMU_DEVICE_BCM59002) || \
    defined(CONFIG_BATTERY_BCM59040) || \
    defined(CONFIG_PMU_DEVICE_BCM59040)
/****************************************************************************
 *
 *   This function initializes the bcm4760_std_battery structure if
 *   we are not using the standard test battery. It may modify the battery_info
 *   member in the bcm4760_battery_device structure to point to a completely
 *   different bcm_batt_info structure if it isn't one of the standard
 *   batteries.
 *
 *****************************************************************************/
static int bcm_batt_platform_init( struct platform_device *pdevice )
{
	return 0;
}

/****************************************************************************
 *
 *   This function does any cleanup in preparation for removal of the
 *   battery device/driver. This is called infrequently at best but the
 *   function is here for completeness.
 *
 *****************************************************************************/

static void bcm_batt_platform_remove( struct platform_device *pdevice )
{
	return;
}

/****************************************************************************
 *
 *   This function is the callback for the ADC function. It is called when
 *   the PMU ADC finishes integration after an ADC request.
 *
 *****************************************************************************/

void bcm_adc_vmbat_handler(struct bcm_adc_request_s *req,
                           unsigned short sample,
                           int error)
{
	struct bcm_ll_adc_req *adc_req;

	mutex_lock(&bcm_vmbat_mutex);
	adc_req = (struct bcm_ll_adc_req *)req->param;
	adc_req->sample = sample;		
	complete(&adc_req->cmp);
	mutex_unlock(&bcm_vmbat_mutex);
}


/****************************************************************************
 *
 *   This function obtains the current battery voltage (VMBAT) using the
 *   PMU ADC. This function will pend the calling task until the ADC completes
 *   integration and returns the voltage.
 *
 *****************************************************************************/
u32 bcm_adc_read(int device, int channel)
{
	struct bcm_ll_adc_req   adc_req;
	u32                     data;

	memset( &adc_req, 0, sizeof( adc_req ) );

	/* Fill in request structure for ADC */
	mutex_lock(&bcm_bat_mutex);
	adc_req.adc.device = device;
	adc_req.adc.channel = channel;
	adc_req.adc.param = (unsigned long int)&adc_req;
	init_completion(&adc_req.cmp);
	adc_req.adc.callback = &bcm_adc_vmbat_handler;

	if (bcm_adc_request(&adc_req.adc) != 0) {
		printk(KERN_ERR PFX "ADC request failed\n");
		mutex_unlock(&bcm_bat_mutex);
		return 0;
	}

	mutex_unlock(&bcm_bat_mutex);
	if (wait_for_completion_timeout(&adc_req.cmp, HZ/2))
		data = (u32)(adc_req.sample);
	else {
		bcm_adc_cancel_request (&adc_req.adc);
		printk (KERN_INFO PFX "ADC request timed out\n");
		data = 0;
	}

	return data; 
}
#endif

#if defined(CONFIG_BATTERY_BCM59040)
/****************************************************************************
 *
 *   This function reads the fuel gauge sample and returns it.
 *
 *****************************************************************************/

static s16 bcm_batt_get_fuel_gauge_sample( void )
{
    short sample;

    PMU_fuelgauge_enable(current_pmu,1);
    msleep(300);
    PMU_fuelgauge_get_sample(current_pmu, &sample, 1);
    PMU_fuelgauge_enable(current_pmu,0);

    return (s16)sample;
}

/****************************************************************************
 *
 *   This function obtains the charge current being used by the PMU battery
 *   charger currently.
 *
 *****************************************************************************/

static u32 bcm_batt_get_charge_current( void )
{
    return (u32)PMU_charger_get_current(current_pmu);
}
#else
#define bcm_batt_get_charge_current NULL
#define bcm_batt_get_fuel_gauge_sample NULL
#endif


static batt_discharge_lut_t bcm4760_std_batt_cap[] = {

	[0] = {
		4106000, 100, 0
	},
	[1] = {
		4031000, 98, 0
	},
	[2] = {
		3879000, 82, 0
	},
	[3] = {
		3796000, 70, 0
	},
	[4] = {
		3699000, 52, 0
	},
	[5] = {
		3650000, 37, 0
	},
	[6] = {
		3601000, 19, 0
	},
	[7] = {
		3552000, 10, 0
	},
	[8] = {
		3527000, 5, 0
	},
	[9] = {
		3495000, 3, 0
	},
	[10] = {
		3457000, 2, 0
	},
	[11] = {
		3392000, 1, 0
	},
	[12] = {
        	3326000, 0, 0
	}
};

static batt_esr_curve_t bcm4760_std_batt_esr[] = {

	[0] = {
		4162300,	259600
	},
	[1] = {
		4096000,	233300
	},
	[2] = {
		4022800,	232800
	},
	[3] = {
		3815800,	243000
	},
	[4] = {
		3609000,	271100
	},
	[5] = {
		3402000,	285700
	},
	[6] = {
		3200000,	305300
	},
};


static struct bcm_batt_info bcm4760_std_battery = {
    .adc = {
	.battery_voltage_resistor_1	= 0,
	.battery_voltage_resistor_2	= 0,
	.charge_current_resistor	= 10000,		/* Resistance in uOhm */
	.reference_voltage		= 0,
	.accuracy			= 0,
	.correction			= 0
    },
    .battery = {
	.discharge_curves		= bcm4760_std_batt_cap,
	.discharge_curve_entries	= ARRAY_SIZE(bcm4760_std_batt_cap),
	.esr_curves			= bcm4760_std_batt_esr,
	.esr_curve_entries		= ARRAY_SIZE(bcm4760_std_batt_esr),
	.Vmax				= 4200,
	.Vmin				= 3200,
	.Vmaxf				= 0,
	.Vminf				= 0
    }
};


static struct bcm_battery_platform_info bcm4760_battery_platinfo= {
	.platform_init		= bcm_batt_platform_init,
	.platform_remove	= bcm_batt_platform_remove,
	.adc_read		= bcm_adc_read,
	.battery_info		= &bcm4760_std_battery,

#if defined(CONFIG_BATTERY_BCM59040)
	.get_charge_current_adc	= bcm_batt_get_charge_current,
	.get_fuel_gauge_sample	= bcm_batt_get_fuel_gauge_sample,
#endif

};

static struct platform_device bcm59040_battery_device = 
{
    .name           = BCM_BATTERY_DRIVER,
    .id             = -1,
    .dev = {
    	.platform_data	= &bcm4760_battery_platinfo,
    },
};

static __init int bcm4760_battery_init(void)
{
	return platform_device_register(&bcm59040_battery_device);
}

module_init(bcm4760_battery_init);
